@@ -28,9 +28,7 @@
  * MT safe
  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "config.h"
 
 #define _GNU_SOURCE		/* For stpcpy */
 
@@ -44,14 +42,19 @@
 #if !defined (HAVE_STRSIGNAL) || !defined(NO_SYS_SIGLIST_DECL)
 #include <signal.h>
 #endif
+
 #include "glib.h"
+#include "gprintf.h"
+#include "gprintfint.h"
+
+#include "galias.h"
 
 #ifdef G_OS_WIN32
 #include <windows.h>
 #endif
 
 /* do not include <unistd.h> in this place since it
- * inteferes with g_strsignal() on some OSes
+ * interferes with g_strsignal() on some OSes
  */
 
 static const guint16 ascii_table_data[256] = {
@@ -74,20 +77,19 @@ static const guint16 ascii_table_data[256] = {
   /* the upper 128 are all zeroes */
 };
 
-#if defined(G_PLATFORM_WIN32) && defined(__GNUC__)
-__declspec(dllexport)
-#endif
 const guint16 * const g_ascii_table = ascii_table_data;
 
 gchar*
 g_strdup (const gchar *str)
 {
   gchar *new_str;
+  gsize length;
 
   if (str)
     {
-      new_str = g_new (char, strlen (str) + 1);
-      strcpy (new_str, str);
+      length = strlen (str) + 1;
+      new_str = g_new (char, length);
+      memcpy (new_str, str, length);
     }
   else
     new_str = NULL;
@@ -179,28 +181,13 @@ g_stpcpy (gchar       *dest,
 
 gchar*
 g_strdup_vprintf (const gchar *format,
-		  va_list      args1)
+		  va_list      args)
 {
-  gchar *buffer;
-#ifdef HAVE_VASPRINTF
-  vasprintf (&buffer, format, args1);
-  if (!g_mem_is_system_malloc ()) 
-    {
-      gchar *buffer1 = g_strdup (buffer);
-      free (buffer);
-      buffer = buffer1;
-    }
-#else
-  va_list args2;
+  gchar *string = NULL;
 
-  G_VA_COPY (args2, args1);
+  g_vasprintf (&string, format, args);
 
-  buffer = g_new (gchar, g_printf_string_upper_bound (format, args1));
-
-  vsprintf (buffer, format, args2);
-  va_end (args2);
-#endif
-  return buffer;
+  return string;
 }
 
 gchar*
@@ -226,7 +213,8 @@ g_strconcat (const gchar *string1, ...)
   gchar	  *concat;
   gchar   *ptr;
 
-  g_return_val_if_fail (string1 != NULL, NULL);
+  if (!string1)
+    return NULL;
 
   l = 1 + strlen (string1);
   va_start (args, string1);
@@ -257,22 +245,22 @@ g_strconcat (const gchar *string1, ...)
 /**
  * g_strtod:
  * @nptr:    the string to convert to a numeric value.
- * @endptr:  if non-NULL, it returns the character after
+ * @endptr:  if non-%NULL, it returns the character after
  *           the last character used in the conversion.
  * 
- * Converts a string to a gdouble value.
+ * Converts a string to a #gdouble value.
  * It calls the standard strtod() function to handle the conversion, but
  * if the string is not completely converted it attempts the conversion
- * again with @g_ascii_strtod, and returns the best match.
+ * again with g_ascii_strtod(), and returns the best match.
  *
- * This function should seldom be used. The normal situation when reading
- * numbers not for human consumption is to use @g_ascii_strtod(). Only when
- * you know that you must expect both locale formated and C formated numbers
+ * This function should seldomly be used. The normal situation when reading
+ * numbers not for human consumption is to use g_ascii_strtod(). Only when
+ * you know that you must expect both locale formatted and C formatted numbers
  * should you use this. Make sure that you don't pass strings such as comma
  * separated lists of values, since the commas may be interpreted as a decimal
  * point in some locales, causing unexpected results.
  * 
- * Return value: the gdouble value.
+ * Return value: the #gdouble value.
  **/
 gdouble
 g_strtod (const gchar *nptr,
@@ -310,32 +298,32 @@ g_strtod (const gchar *nptr,
 /**
  * g_ascii_strtod:
  * @nptr:    the string to convert to a numeric value.
- * @endptr:  if non-NULL, it returns the character after
+ * @endptr:  if non-%NULL, it returns the character after
  *           the last character used in the conversion.
  * 
- * Converts a string to a gdouble value.
+ * Converts a string to a #gdouble value.
  * This function behaves like the standard strtod() function
  * does in the C locale. It does this without actually
  * changing the current locale, since that would not be
  * thread-safe.
  *
  * This function is typically used when reading configuration
- * files or other non-user input that should be locale dependent.
+ * files or other non-user input that should be locale independent.
  * To handle input from the user you should normally use the
- * locale-sensitive system strtod function.
+ * locale-sensitive system strtod() function.
  *
- * To convert from a string to double in a locale-insensitive
- * way, use @g_ascii_dtostr.
+ * To convert from a #gdouble to a string in a locale-insensitive
+ * way, use g_ascii_dtostr().
  *
- * If the correct value would cause overflow, plus or minus HUGE_VAL
- * is returned (according to the sign of the value), and ERANGE is
- * stored in errno. If the correct value would cause underflow,
- * zero is returned and ERANGE is stored in errno.
+ * If the correct value would cause overflow, plus or minus %HUGE_VAL
+ * is returned (according to the sign of the value), and %ERANGE is
+ * stored in %errno. If the correct value would cause underflow,
+ * zero is returned and %ERANGE is stored in %errno.
  * 
- * This function resets errno before calling strtod() so that
+ * This function resets %errno before calling strtod() so that
  * you can reliably detect overflow and underflow.
  *
- * Return value: the gdouble value.
+ * Return value: the #gdouble value.
  **/
 gdouble
 g_ascii_strtod (const gchar *nptr,
@@ -348,6 +336,7 @@ g_ascii_strtod (const gchar *nptr,
   int decimal_point_len;
   const char *p, *decimal_point_pos;
   const char *end = NULL; /* Silence gcc */
+  int strtod_errno;
 
   g_return_val_if_fail (nptr != NULL, 0);
 
@@ -360,7 +349,9 @@ g_ascii_strtod (const gchar *nptr,
   g_assert (decimal_point_len != 0);
   
   decimal_point_pos = NULL;
-  if (decimal_point[0] != '.' ||
+  end = NULL;
+
+  if (decimal_point[0] != '.' || 
       decimal_point[1] != 0)
     {
       p = nptr;
@@ -372,7 +363,7 @@ g_ascii_strtod (const gchar *nptr,
       if (*p == '+' || *p == '-')
 	p++;
       
-      if (p[0] == '0' &&
+      if (p[0] == '0' && 
 	  (p[1] == 'x' || p[1] == 'X'))
 	{
 	  p += 2;
@@ -382,49 +373,43 @@ g_ascii_strtod (const gchar *nptr,
 	    p++;
 	  
 	  if (*p == '.')
-	    {
-	      decimal_point_pos = p++;
+	    decimal_point_pos = p++;
 	      
-	      while (g_ascii_isxdigit (*p))
-		p++;
-	      
-	      if (*p == 'p' || *p == 'P')
-		p++;
-	      if (*p == '+' || *p == '-')
-		p++;
-	      while (g_ascii_isdigit (*p))
-		p++;
-	      end = p;
-	    }
+	  while (g_ascii_isxdigit (*p))
+	    p++;
+	  
+	  if (*p == 'p' || *p == 'P')
+	    p++;
+	  if (*p == '+' || *p == '-')
+	    p++;
+	  while (g_ascii_isdigit (*p))
+	    p++;
+
+	  end = p;
 	}
-      else
+      else if (g_ascii_isdigit (*p) || *p == '.')
 	{
 	  while (g_ascii_isdigit (*p))
 	    p++;
 	  
 	  if (*p == '.')
-	    {
-	      decimal_point_pos = p++;
-	      
-	      while (g_ascii_isdigit (*p))
-		p++;
-	      
-	      if (*p == 'e' || *p == 'E')
-		p++;
-	      if (*p == '+' || *p == '-')
-		p++;
-	      while (g_ascii_isdigit (*p))
-		p++;
-	      end = p;
-	    }
+	    decimal_point_pos = p++;
+	  
+	  while (g_ascii_isdigit (*p))
+	    p++;
+	  
+	  if (*p == 'e' || *p == 'E')
+	    p++;
+	  if (*p == '+' || *p == '-')
+	    p++;
+	  while (g_ascii_isdigit (*p))
+	    p++;
+
+	  end = p;
 	}
       /* For the other cases, we need not convert the decimal point */
     }
 
-  /* Set errno to zero, so that we can distinguish zero results
-     and underflows */
-  errno = 0;
-  
   if (decimal_point_pos)
     {
       char *copy, *c;
@@ -441,11 +426,13 @@ g_ascii_strtod (const gchar *nptr,
       c += end - (decimal_point_pos + 1);
       *c = 0;
 
+      errno = 0;
       val = strtod (copy, &fail_pos);
+      strtod_errno = errno;
 
       if (fail_pos)
 	{
-	  if (fail_pos > decimal_point_pos)
+	  if (fail_pos - copy > decimal_point_pos - nptr)
 	    fail_pos = (char *)nptr + (fail_pos - copy) - (decimal_point_len - 1);
 	  else
 	    fail_pos = (char *)nptr + (fail_pos - copy);
@@ -454,26 +441,52 @@ g_ascii_strtod (const gchar *nptr,
       g_free (copy);
 	  
     }
+  else if (end)
+    {
+      char *copy;
+      
+      copy = g_malloc (end - (char *)nptr + 1);
+      memcpy (copy, nptr, end - nptr);
+      *(copy + (end - (char *)nptr)) = 0;
+      
+      errno = 0;
+      val = strtod (copy, &fail_pos);
+      strtod_errno = errno;
+
+      if (fail_pos)
+	{
+	  fail_pos = (char *)nptr + (fail_pos - copy);
+	}
+      
+      g_free (copy);
+    }
   else
-    val = strtod (nptr, &fail_pos);
+    {
+      errno = 0;
+      val = strtod (nptr, &fail_pos);
+      strtod_errno = errno;
+    }
 
   if (endptr)
     *endptr = fail_pos;
-  
+
+  errno = strtod_errno;
+
   return val;
 }
 
+
 /**
  * g_ascii_dtostr:
  * @buffer: A buffer to place the resulting string in
  * @buf_len: The length of the buffer.
- * @d: The double to convert
+ * @d: The #gdouble to convert
  *
- * Converts a double to a string, using the '.' as
- * decimal_point. 
+ * Converts a #gdouble to a string, using the '.' as
+ * decimal point. 
  * 
  * This functions generates enough precision that converting
- * the string back using @g_strtod gives the same machine-number
+ * the string back using g_ascii_strtod() gives the same machine-number
  * (on machines with IEEE compatible 64bit doubles). It is
  * guaranteed that the size of the resulting string will never
  * be larger than @G_ASCII_DTOSTR_BUF_SIZE bytes.
@@ -492,17 +505,17 @@ g_ascii_dtostr (gchar       *buffer,
  * g_ascii_formatd:
  * @buffer: A buffer to place the resulting string in
  * @buf_len: The length of the buffer.
- * @format: The printf-style format to use for the
+ * @format: The printf()-style format to use for the
  *          code to use for converting. 
- * @d: The double to convert
+ * @d: The #gdouble to convert
  *
- * Converts a double to a string, using the '.' as
- * decimal_point. To format the number you pass in
- * a printf-style formating string. Allowed conversion
- * specifiers are eEfFgG. 
+ * Converts a #gdouble to a string, using the '.' as
+ * decimal point. To format the number you pass in
+ * a printf()-style format string. Allowed conversion
+ * specifiers are 'e', 'E', 'f', 'F', 'g' and 'G'. 
  * 
  * If you just want to want to serialize the value into a
- * string, use @g_ascii_dtostr.
+ * string, use g_ascii_dtostr().
  *
  * Return value: The pointer to the buffer with the converted string.
  **/
@@ -542,7 +555,7 @@ g_ascii_formatd (gchar       *buffer,
     return NULL;
 
       
-  g_snprintf (buffer, buf_len, format, d);
+  _g_snprintf (buffer, buf_len, format, d);
 
   locale_data = localeconv ();
   decimal_point = locale_data->decimal_point;
@@ -578,15 +591,191 @@ g_ascii_formatd (gchar       *buffer,
   return buffer;
 }
 
+/**
+ * g_ascii_strtoull:
+ * @nptr:    the string to convert to a numeric value.
+ * @endptr:  if non-%NULL, it returns the character after
+ *           the last character used in the conversion.
+ * @base:    to be used for the conversion, 2..36 or 0
+ *
+ * Converts a string to a #guint64 value.
+ * This function behaves like the standard strtoull() function
+ * does in the C locale. It does this without actually
+ * changing the current locale, since that would not be
+ * thread-safe.
+ *
+ * This function is typically used when reading configuration
+ * files or other non-user input that should be locale independent.
+ * To handle input from the user you should normally use the
+ * locale-sensitive system strtoull() function.
+ *
+ * If the correct value would cause overflow, %G_MAXUINT64
+ * is returned, and %ERANGE is stored in %errno.
+ *
+ * Return value: the #guint64 value.
+ *
+ * Since: 2.2
+ **/
+guint64
+g_ascii_strtoull (const gchar *nptr,
+		  gchar      **endptr,
+		  guint        base)
+{
+  /* this code is based on on the strtol(3) code from GNU libc released under
+   * the GNU Lesser General Public License.
+   *
+   * Copyright (C) 1991,92,94,95,96,97,98,99,2000,01,02
+   *        Free Software Foundation, Inc.
+   */
+#define ISSPACE(c)		((c) == ' ' || (c) == '\f' || (c) == '\n' || \
+				 (c) == '\r' || (c) == '\t' || (c) == '\v')
+#define ISUPPER(c)		((c) >= 'A' && (c) <= 'Z')
+#define ISLOWER(c)		((c) >= 'a' && (c) <= 'z')
+#define ISALPHA(c)		(ISUPPER (c) || ISLOWER (c))
+#define	TOUPPER(c)		(ISLOWER (c) ? (c) - 'a' + 'A' : (c))
+#define	TOLOWER(c)		(ISUPPER (c) ? (c) - 'A' + 'a' : (c))
+  gboolean negative, overflow;
+  guint64 cutoff;
+  guint64 cutlim;
+  guint64 ui64;
+  const gchar *s, *save;
+  guchar c;
+  
+  g_return_val_if_fail (nptr != NULL, 0);
+  
+  if (base == 1 || base > 36)
+    {
+      errno = EINVAL;
+      return 0;
+    }
+  
+  save = s = nptr;
+  
+  /* Skip white space.  */
+  while (ISSPACE (*s))
+    ++s;
+  if (!*s)
+    goto noconv;
+  
+  /* Check for a sign.  */
+  negative = FALSE;
+  if (*s == '-')
+    {
+      negative = TRUE;
+      ++s;
+    }
+  else if (*s == '+')
+    ++s;
+  
+  /* Recognize number prefix and if BASE is zero, figure it out ourselves.  */
+  if (*s == '0')
+    {
+      if ((base == 0 || base == 16) && TOUPPER (s[1]) == 'X')
+	{
+	  s += 2;
+	  base = 16;
+	}
+      else if (base == 0)
+	base = 8;
+    }
+  else if (base == 0)
+    base = 10;
+  
+  /* Save the pointer so we can check later if anything happened.  */
+  save = s;
+  cutoff = G_MAXUINT64 / base;
+  cutlim = G_MAXUINT64 % base;
+  
+  overflow = FALSE;
+  ui64 = 0;
+  c = *s;
+  for (; c; c = *++s)
+    {
+      if (c >= '0' && c <= '9')
+	c -= '0';
+      else if (ISALPHA (c))
+	c = TOUPPER (c) - 'A' + 10;
+      else
+	break;
+      if (c >= base)
+	break;
+      /* Check for overflow.  */
+      if (ui64 > cutoff || (ui64 == cutoff && c > cutlim))
+	overflow = TRUE;
+      else
+	{
+	  ui64 *= base;
+	  ui64 += c;
+	}
+    }
+  
+  /* Check if anything actually happened.  */
+  if (s == save)
+    goto noconv;
+  
+  /* Store in ENDPTR the address of one character
+     past the last character we converted.  */
+  if (endptr)
+    *endptr = (gchar*) s;
+  
+  if (overflow)
+    {
+      errno = ERANGE;
+      return G_MAXUINT64;
+    }
+  
+  /* Return the result of the appropriate sign.  */
+  return negative ? -ui64 : ui64;
+  
+ noconv:
+  /* We must handle a special case here: the base is 0 or 16 and the
+     first two characters are '0' and 'x', but the rest are no
+     hexadecimal digits.  This is no error case.  We return 0 and
+     ENDPTR points to the `x`.  */
+  if (endptr)
+    {
+      if (save - nptr >= 2 && TOUPPER (save[-1]) == 'X'
+	  && save[-2] == '0')
+	*endptr = (gchar*) &save[-1];
+      else
+	/*  There was no number to convert.  */
+	*endptr = (gchar*) nptr;
+    }
+  return 0;
+}
+
 
 G_CONST_RETURN gchar*
 g_strerror (gint errnum)
 {
   static GStaticPrivate msg_private = G_STATIC_PRIVATE_INIT;
   char *msg;
+  int saved_errno = errno;
 
 #ifdef HAVE_STRERROR
-  return strerror (errnum);
+  const char *msg_locale;
+
+  msg_locale = strerror (errnum);
+  if (g_get_charset (NULL))
+    {
+      errno = saved_errno;
+      return msg_locale;
+    }
+  else
+    {
+      gchar *msg_utf8 = g_locale_to_utf8 (msg_locale, -1, NULL, NULL, NULL);
+      if (msg_utf8)
+	{
+	  /* Stick in the quark table so that we can return a static result
+	   */
+	  GQuark msg_quark = g_quark_from_string (msg_utf8);
+	  g_free (msg_utf8);
+	  
+	  msg_utf8 = (gchar *) g_quark_to_string (msg_quark);
+	  errno = saved_errno;
+	  return msg_utf8;
+	}
+    }
 #elif NO_SYS_ERRLIST
   switch (errnum)
     {
@@ -1017,8 +1206,9 @@ g_strerror (gint errnum)
       g_static_private_set (&msg_private, msg, g_free);
     }
 
-  sprintf (msg, "unknown error (%d)", errnum);
+  _g_sprintf (msg, "unknown error (%d)", errnum);
 
+  errno = saved_errno;
   return msg;
 }
 
@@ -1029,13 +1219,30 @@ g_strsignal (gint signum)
   char *msg;
 
 #ifdef HAVE_STRSIGNAL
+  const char *msg_locale;
+  
 #if defined(G_OS_BEOS) || defined(G_WITH_CYGWIN)
 extern const char *strsignal(int);
 #else
   /* this is declared differently (const) in string.h on BeOS */
   extern char *strsignal (int sig);
 #endif /* !G_OS_BEOS && !G_WITH_CYGWIN */
-  return strsignal (signum);
+  msg_locale = strsignal (signum);
+  if (g_get_charset (NULL))
+    return msg_locale;
+  else
+    {
+      gchar *msg_utf8 = g_locale_to_utf8 (msg_locale, -1, NULL, NULL, NULL);
+      if (msg_utf8)
+	{
+	  /* Stick in the quark table so that we can return a static result
+	   */
+	  GQuark msg_quark = g_quark_from_string (msg_utf8);
+	  g_free (msg_utf8);
+	  
+	  return g_quark_to_string (msg_quark);
+	}
+    }
 #elif NO_SYS_SIGLIST
   switch (signum)
     {
@@ -1149,7 +1356,7 @@ extern const char *strsignal(int);
       g_static_private_set (&msg_private, msg, g_free);
     }
 
-  sprintf (msg, "unknown signal (%d)", signum);
+  _g_sprintf (msg, "unknown signal (%d)", signum);
   
   return msg;
 }
@@ -1341,6 +1548,18 @@ g_ascii_strup (const gchar *str,
   return result;
 }
 
+/**
+ * g_strdown:
+ * @string: the string to convert.
+ * 
+ * Converts a string to lower case.  
+ * 
+ * Return value: the string 
+ *
+ * Deprecated: This function is totally broken for the reasons discussed in 
+ * the g_strncasecmp() docs - use g_ascii_strdown() or g_utf8_strdown() 
+ * instead.
+ **/
 gchar*
 g_strdown (gchar *string)
 {
@@ -1360,6 +1579,17 @@ g_strdown (gchar *string)
   return (gchar *) string;
 }
 
+/**
+ * g_strup:
+ * @string: the string to convert.
+ * 
+ * Converts a string to upper case. 
+ * 
+ * Return value: the string
+ *
+ * Deprecated: This function is totally broken for the reasons discussed in 
+ * the g_strncasecmp() docs - use g_ascii_strup() or g_utf8_strup() instead.
+ **/
 gchar*
 g_strup (gchar *string)
 {
@@ -1522,8 +1752,8 @@ g_ascii_strcasecmp (const gchar *s1,
 
   while (*s1 && *s2)
     {
-      c1 = (gint)(guchar) g_ascii_tolower (*s1);
-      c2 = (gint)(guchar) g_ascii_tolower (*s2);
+      c1 = (gint)(guchar) TOLOWER (*s1);
+      c2 = (gint)(guchar) TOLOWER (*s2);
       if (c1 != c2)
 	return (c1 - c2);
       s1++; s2++;
@@ -1563,8 +1793,8 @@ g_ascii_strncasecmp (const gchar *s1,
   while (n && *s1 && *s2)
     {
       n -= 1;
-      c1 = (gint)(guchar) g_ascii_tolower (*s1);
-      c2 = (gint)(guchar) g_ascii_tolower (*s2);
+      c1 = (gint)(guchar) TOLOWER (*s1);
+      c2 = (gint)(guchar) TOLOWER (*s2);
       if (c1 != c2)
 	return (c1 - c2);
       s1++; s2++;
@@ -1576,6 +1806,20 @@ g_ascii_strncasecmp (const gchar *s1,
     return 0;
 }
 
+/**
+ * g_strcasecmp:
+ * @s1: a string.
+ * @s2: a string to compare with @s1.
+ * 
+ * A case-insensitive string comparison, corresponding to the standard
+ * strcasecmp() function on platforms which support it.
+ *
+ * Return value: 0 if the strings match, a negative value if @s1 &lt; @s2, 
+ *   or a positive value if @s1 &gt; @s2.
+ *
+ * Deprecated: See g_strncasecmp() for a discussion of why this function is 
+ *   deprecated and how to replace it.
+ **/
 gint
 g_strcasecmp (const gchar *s1,
 	      const gchar *s2)
@@ -1607,10 +1851,38 @@ g_strcasecmp (const gchar *s1,
 #endif
 }
 
+/**
+ * g_strncasecmp:
+ * @s1: a string.
+ * @s2: a string to compare with @s1.
+ * @n: the maximum number of characters to compare.
+ * 
+ * A case-insensitive string comparison, corresponding to the standard
+ * strncasecmp() function on platforms which support it.
+ * It is similar to g_strcasecmp() except it only compares the first @n 
+ * characters of the strings.
+ * 
+ * Return value: 0 if the strings match, a negative value if @s1 &lt; @s2, 
+ *   or a positive value if @s1 &gt; @s2.
+ *
+ * Deprecated: The problem with g_strncasecmp() is that it does the 
+ * comparison by calling toupper()/tolower(). These functions are
+ * locale-specific and operate on single bytes. However, it is impossible
+ * to handle things correctly from an I18N standpoint by operating on
+ * bytes, since characters may be multibyte. Thus g_strncasecmp() is
+ * broken if your string is guaranteed to be ASCII, since it's
+ * locale-sensitive, and it's broken if your string is localized, since
+ * it doesn't work on many encodings at all, including UTF-8, EUC-JP,
+ * etc.
+ *
+ * There are therefore two replacement functions: g_ascii_strncasecmp(),
+ * which only works on ASCII and is not locale-sensitive, and
+ * g_utf8_casefold(), which is good for case-insensitive sorting of UTF-8.
+ **/
 gint
 g_strncasecmp (const gchar *s1,
 	       const gchar *s2,
-	       gsize n)     
+	       guint n)     
 {
 #ifdef HAVE_STRNCASECMP
   return strncasecmp (s1, s2, n);
@@ -1835,16 +2107,18 @@ g_strchug (gchar *string)
 gchar*
 g_strchomp (gchar *string)
 {
-  gchar *s;
+  gsize len;
 
   g_return_val_if_fail (string != NULL, NULL);
 
-  if (!*string)
-    return string;
-
-  for (s = string + strlen (string) - 1; s >= string && g_ascii_isspace ((guchar)*s); 
-       s--)
-    *s = '\0';
+  len = strlen (string);
+  while (len--)
+    {
+      if (g_ascii_isspace ((guchar) string[len]))
+	string[len] = '\0';
+      else
+	break;
+    }
 
   return string;
 }
@@ -1927,6 +2201,113 @@ g_strsplit (const gchar *string,
   return str_array;
 }
 
+/**
+ * g_strsplit_set:
+ * @string: The string to be tokenized
+ * @delimiters: A nul-terminated string containing bytes that are used
+ *              to split the string.
+ * @max_tokens: The maximum number of tokens to split @string into. 
+ *              If this is less than 1, the string is split completely
+ * 
+ * Splits @string into a number of tokens not containing any of the characters
+ * in @delimiter. A token is the (possibly empty) longest string that does not
+ * contain any of the characters in @delimiters. If @max_tokens is reached, the
+ * remainder is appended to the last token.
+ *
+ * For example the result of g_strsplit_set ("abc:def/ghi", ":/", -1) is a
+ * %NULL-terminated vector containing the three strings "abc", "def", 
+ * and "ghi".
+ *
+ * The result if g_strsplit_set (":def/ghi:", ":/", -1) is a %NULL-terminated
+ * vector containing the four strings "", "def", "ghi", and "".
+ * 
+ * As a special case, the result of splitting the empty string "" is an empty
+ * vector, not a vector containing a single string. The reason for this
+ * special case is that being able to represent a empty vector is typically
+ * more useful than consistent handling of empty elements. If you do need
+ * to represent empty elements, you'll need to check for the empty string
+ * before calling g_strsplit_set().
+ *
+ * Note that this function works on bytes not characters, so it can't be used 
+ * to delimit UTF-8 strings for anything but ASCII characters.
+ * 
+ * Return value: a newly-allocated %NULL-terminated array of strings. Use 
+ *    g_strfreev() to free it.
+ * 
+ * Since: 2.4
+ **/
+gchar **
+g_strsplit_set (const gchar *string,
+	        const gchar *delimiters,
+	        gint         max_tokens)
+{
+  gboolean delim_table[256];
+  GSList *tokens, *list;
+  gint n_tokens;
+  const gchar *s;
+  const gchar *current;
+  gchar *token;
+  gchar **result;
+  
+  g_return_val_if_fail (string != NULL, NULL);
+  g_return_val_if_fail (delimiters != NULL, NULL);
+
+  if (max_tokens < 1)
+    max_tokens = G_MAXINT;
+
+  if (*string == '\0')
+    {
+      result = g_new (char *, 1);
+      result[0] = NULL;
+      return result;
+    }
+  
+  memset (delim_table, FALSE, sizeof (delim_table));
+  for (s = delimiters; *s != '\0'; ++s)
+    delim_table[*(guchar *)s] = TRUE;
+
+  tokens = NULL;
+  n_tokens = 0;
+
+  s = current = string;
+  while (*s != '\0')
+    {
+      if (delim_table[*(guchar *)s] && n_tokens + 1 < max_tokens)
+	{
+	  gchar *token;
+
+	  token = g_strndup (current, s - current);
+	  tokens = g_slist_prepend (tokens, token);
+	  ++n_tokens;
+
+	  current = s + 1;
+	}
+      
+      ++s;
+    }
+
+  token = g_strndup (current, s - current);
+  tokens = g_slist_prepend (tokens, token);
+  ++n_tokens;
+
+  result = g_new (gchar *, n_tokens + 1);
+
+  result[n_tokens] = NULL;
+  for (list = tokens; list != NULL; list = list->next)
+    result[--n_tokens] = list->data;
+
+  g_slist_free (tokens);
+  
+  return result;
+}
+
+/**
+ * g_strfreev:
+ * @str_array: a %NULL-terminated array of strings to free.
+
+ * Frees a %NULL-terminated array of strings, and the array itself.
+ * If called on a %NULL value, g_strfreev() simply returns. 
+ **/
 void
 g_strfreev (gchar **str_array)
 {
@@ -2235,3 +2616,118 @@ g_strrstr_len (const gchar *haystack,
 }
 
 
+/**
+ * g_str_has_suffix:
+ * @str: a nul-terminated string.
+ * @suffix: the nul-terminated suffix to look for.
+ *
+ * Looks whether the string @str ends with @suffix.
+ *
+ * Return value: %TRUE if @str end with @suffix, %FALSE otherwise.
+ *
+ * Since: 2.2
+ **/
+gboolean
+g_str_has_suffix (const gchar  *str,
+		  const gchar  *suffix)
+{
+  int str_len;
+  int suffix_len;
+  
+  g_return_val_if_fail (str != NULL, FALSE);
+  g_return_val_if_fail (suffix != NULL, FALSE);
+
+  str_len = strlen (str);
+  suffix_len = strlen (suffix);
+
+  if (str_len < suffix_len)
+    return FALSE;
+
+  return strcmp (str + str_len - suffix_len, suffix) == 0;
+}
+
+/**
+ * g_str_has_prefix:
+ * @str: a nul-terminated string.
+ * @prefix: the nul-terminated prefix to look for.
+ *
+ * Looks whether the string @str begins with @prefix.
+ *
+ * Return value: %TRUE if @str begins with @prefix, %FALSE otherwise.
+ *
+ * Since: 2.2
+ **/
+gboolean
+g_str_has_prefix (const gchar  *str,
+		  const gchar  *prefix)
+{
+  int str_len;
+  int prefix_len;
+  
+  g_return_val_if_fail (str != NULL, FALSE);
+  g_return_val_if_fail (prefix != NULL, FALSE);
+
+  str_len = strlen (str);
+  prefix_len = strlen (prefix);
+
+  if (str_len < prefix_len)
+    return FALSE;
+  
+  return strncmp (str, prefix, prefix_len) == 0;
+}
+
+
+/**
+ * g_strip_context:
+ * @msgid: a string
+ * @msgval: another string
+ * 
+ * An auxiliary function for gettext() support (see Q_()).
+ * 
+ * Return value: @msgval, unless @msgval is identical to @msgid and contains
+ *   a '|' character, in which case a pointer to the substring of msgid after
+ *   the first '|' character is returned. 
+ *
+ * Since: 2.4
+ **/
+G_CONST_RETURN gchar *
+g_strip_context  (const gchar *msgid, 
+		  const gchar *msgval)
+{
+  if (msgval == msgid)
+    {
+      const char *c = strchr (msgid, '|');
+      if (c != NULL)
+	return c + 1;
+    }
+  
+  return msgval;
+}
+
+
+/**
+ * g_strv_length:
+ * @str_array: a %NULL-terminated array of strings.
+ * 
+ * Returns the length of the given %NULL-terminated 
+ * string array @str_array.
+ * 
+ * Return value: length of @str_array.
+ *
+ * Since: 2.6
+ **/
+guint
+g_strv_length (gchar **str_array)
+{
+  guint i = 0;
+
+  g_return_val_if_fail (str_array != NULL, 0);
+
+  while (str_array[i])
+    ++i;
+
+  return i;
+}
+
+#define __G_STRFUNCS_C__
+#include "galiasdef.c"
