@@ -24,12 +24,19 @@
  * MT safe
  */
 
+#include <config.h>
+
 #include        "gsignal.h"
+
 #include        "gbsearcharray.h"
 #include        "gvaluecollector.h"
 #include	"gvaluetypes.h"
 #include	"gboxed.h"
+
+#include	"gobjectalias.h"
+
 #include	<string.h> 
+#include	<signal.h>
 
 
 /* pre allocation configurations
@@ -37,7 +44,7 @@
 #define	MAX_STACK_VALUES	(16)
 #define HANDLER_PRE_ALLOC       (48)
 
-#define REPORT_BUG      "please report occourance circumstances to gtk-devel-list@gnome.org"
+#define REPORT_BUG      "please report occurrence circumstances to gtk-devel-list@gnome.org"
 #ifdef	G_ENABLE_DEBUG
 #define IF_DEBUG(debug_type, cond)	if ((_g_type_debug_flags & G_TYPE_DEBUG_ ## debug_type) || cond)
 static volatile gpointer g_trace_instance_signals = NULL;
@@ -147,6 +154,7 @@ static	      gboolean		signal_emit_unlocked_R	(SignalNode	 *node,
 							 gpointer	  instance,
 							 GValue		 *return_value,
 							 const GValue	 *instance_and_params);
+static const gchar *            type_debug_name         (GType            type);
 
 
 /* --- structures --- */
@@ -171,6 +179,7 @@ struct _SignalNode
   guint              destroyed : 1;
   
   /* reinitializable portion */
+  guint		     test_class_offset : 12;
   guint              flags : 8;
   guint              n_params : 8;
   GType		    *param_types; /* mangled with G_SIGNAL_TYPE_STATIC_SCOPE flag */
@@ -180,6 +189,8 @@ struct _SignalNode
   GSignalCMarshaller c_marshaller;
   GHookList         *emission_hooks;
 };
+#define	MAX_TEST_CLASS_OFFSET	(4096)	/* 2^12, 12 bits for test_class_offset */
+#define	TEST_CLASS_MAGIC	(1)	/* indicates NULL class closure, candidate for NOP optimization */
 
 struct _SignalKey
 {
@@ -201,7 +212,10 @@ struct _HandlerList
 {
   guint    signal_id;
   Handler *handlers;
+  Handler *tail_before;  /* normal signal handlers are appended here  */
+  Handler *tail_after;   /* CONNECT_AFTER handlers are appended here  */
 };
+
 struct _Handler
 {
   gulong        sequential_number;
@@ -234,19 +248,27 @@ typedef struct
 
 /* --- variables --- */
 static GBSearchArray *g_signal_key_bsa = NULL;
-static GBSearchConfig g_signal_key_bconfig = G_STATIC_BCONFIG (sizeof (SignalKey),
-							       signal_key_cmp,
-							       G_BSEARCH_ARRAY_ALIGN_POWER2);
-static GBSearchConfig g_signal_hlbsa_bconfig = G_STATIC_BCONFIG (sizeof (HandlerList),
-								 handler_lists_cmp,
-								 G_BSEARCH_ARRAY_DEFER_SHRINK);
-static GBSearchConfig g_class_closure_bconfig = G_STATIC_BCONFIG (sizeof (ClassClosure),
-								  class_closures_cmp,
-								  0);
+static const GBSearchConfig g_signal_key_bconfig = {
+  sizeof (SignalKey),
+  signal_key_cmp,
+  G_BSEARCH_ARRAY_ALIGN_POWER2,
+};
+static GBSearchConfig g_signal_hlbsa_bconfig = {
+  sizeof (HandlerList),
+  handler_lists_cmp,
+  0,
+};
+static GBSearchConfig g_class_closure_bconfig = {
+  sizeof (ClassClosure),
+  class_closures_cmp,
+  0,
+};
 static GHashTable    *g_handler_list_bsa_ht = NULL;
 static Emission      *g_recursive_emissions = NULL;
 static Emission      *g_restart_emissions = NULL;
+#ifndef DISABLE_MEM_POOLS
 static GTrashStack   *g_handler_ts = NULL;
+#endif
 static gulong         g_handler_sequential_number = 1;
 G_LOCK_DEFINE_STATIC (g_signal_mutex);
 #define	SIGNAL_LOCK()		G_LOCK (g_signal_mutex)
@@ -278,7 +300,7 @@ signal_id_lookup (GQuark quark,
 
   key.quark = quark;
 
-  /* try looking up signals for this type and its anchestors */
+  /* try looking up signals for this type and its ancestors */
   do
     {
       SignalKey *signal_key;
@@ -339,18 +361,20 @@ handler_list_ensure (guint    signal_id,
   HandlerList key;
   
   key.signal_id = signal_id;
-  key.handlers = NULL;
+  key.handlers    = NULL;
+  key.tail_before = NULL;
+  key.tail_after  = NULL;
   if (!hlbsa)
     {
-      hlbsa = g_bsearch_array_new (&g_signal_hlbsa_bconfig);
-      hlbsa = g_bsearch_array_insert (hlbsa, &g_signal_hlbsa_bconfig, &key, FALSE);
+      hlbsa = g_bsearch_array_create (&g_signal_hlbsa_bconfig);
+      hlbsa = g_bsearch_array_insert (hlbsa, &g_signal_hlbsa_bconfig, &key);
       g_hash_table_insert (g_handler_list_bsa_ht, instance, hlbsa);
     }
   else
     {
       GBSearchArray *o = hlbsa;
 
-      hlbsa = g_bsearch_array_insert (o, &g_signal_hlbsa_bconfig, &key, FALSE);
+      hlbsa = g_bsearch_array_insert (o, &g_signal_hlbsa_bconfig, &key);
       if (hlbsa != o)
 	g_hash_table_insert (g_handler_list_bsa_ht, instance, hlbsa);
     }
@@ -561,16 +585,45 @@ handler_unref_R (guint    signal_id,
   handler->ref_count -= 1;
   if (!handler->ref_count)
     {
+      HandlerList *hlist = NULL;
+
       if (handler->next)
         handler->next->prev = handler->prev;
-      if (handler->prev)	/* watch out for g_signal_handlers_destroy()! */
+      if (handler->prev)    /* watch out for g_signal_handlers_destroy()! */
         handler->prev->next = handler->next;
       else
         {
-          HandlerList *hlist = handler_list_lookup (signal_id, instance);
-          
+          hlist = handler_list_lookup (signal_id, instance);
           hlist->handlers = handler->next;
         }
+
+      if (instance)
+        {
+          /*  check if we are removing the handler pointed to by tail_before  */
+          if (!handler->after && (!handler->next || handler->next->after))
+            {
+              if (!hlist)
+                hlist = handler_list_lookup (signal_id, instance);
+              if (hlist)
+                {
+                  g_assert (hlist->tail_before == handler); /* paranoid */
+                  hlist->tail_before = handler->prev;
+                }
+            }
+
+          /*  check if we are removing the handler pointed to by tail_after  */
+          if (!handler->next)
+            {
+              if (!hlist)
+                hlist = handler_list_lookup (signal_id, instance);
+              if (hlist)
+                {
+                  g_assert (hlist->tail_after == handler); /* paranoid */
+                  hlist->tail_after = handler->prev;
+                }
+            }
+        }
+
       SIGNAL_UNLOCK ();
       g_closure_unref (handler->closure);
       SIGNAL_LOCK ();
@@ -589,29 +642,38 @@ handler_insert (guint    signal_id,
   
   hlist = handler_list_ensure (signal_id, instance);
   if (!hlist->handlers)
-    hlist->handlers = handler;
-  else if (hlist->handlers->after && !handler->after)
     {
-      handler->next = hlist->handlers;
-      hlist->handlers->prev = handler;
       hlist->handlers = handler;
+      if (!handler->after)
+        hlist->tail_before = handler;
+    }
+  else if (handler->after)
+    {
+      handler->prev = hlist->tail_after;
+      hlist->tail_after->next = handler;
     }
   else
     {
-      Handler *tmp = hlist->handlers;
-      
-      if (handler->after)
-        while (tmp->next)
-          tmp = tmp->next;
-      else
-        while (tmp->next && !tmp->next->after)
-          tmp = tmp->next;
-      if (tmp->next)
-        tmp->next->prev = handler;
-      handler->next = tmp->next;
-      handler->prev = tmp;
-      tmp->next = handler;
+      if (hlist->tail_before)
+        {
+          handler->next = hlist->tail_before->next;
+          if (handler->next)
+            handler->next->prev = handler;
+          handler->prev = hlist->tail_before;
+          hlist->tail_before->next = handler;
+        }
+      else /* insert !after handler into a list of only after handlers */
+        {
+          handler->next = hlist->handlers;
+          if (handler->next)
+            handler->next->prev = handler;
+          hlist->handlers = handler;
+        }
+      hlist->tail_before = handler;
     }
+
+  if (!handler->next)
+    hlist->tail_after = handler;
 }
 
 static inline void
@@ -694,7 +756,7 @@ signal_key_cmp (gconstpointer node1,
 }
 
 void
-g_signal_init (void) /* sync with gtype.c */
+g_signal_init (void)
 {
   SIGNAL_LOCK ();
   if (!g_n_signal_nodes)
@@ -704,7 +766,7 @@ g_signal_init (void) /* sync with gtype.c */
       
       /* setup handler list binary searchable array hash table (in german, that'd be one word ;) */
       g_handler_list_bsa_ht = g_hash_table_new (g_direct_hash, NULL);
-      g_signal_key_bsa = g_bsearch_array_new (&g_signal_key_bconfig);
+      g_signal_key_bsa = g_bsearch_array_create (&g_signal_key_bconfig);
       
       /* invalid (0) signal_id */
       g_n_signal_nodes = 1;
@@ -729,7 +791,7 @@ _g_signals_destroy (GType itype)
           if (node->destroyed)
             g_warning (G_STRLOC ": signal \"%s\" of type `%s' already destroyed",
                        node->name,
-                       g_type_name (node->itype));
+                       type_debug_name (node->itype));
           else
 	    signal_destroy_R (node);
         }
@@ -809,12 +871,18 @@ g_signal_add_emission_hook (guint               signal_id,
 
   SIGNAL_LOCK ();
   node = LOOKUP_SIGNAL_NODE (signal_id);
-  if (!node || node->destroyed || (node->flags & G_SIGNAL_NO_HOOKS))
+  if (!node || node->destroyed)
     {
       g_warning ("%s: invalid signal id `%u'", G_STRLOC, signal_id);
       SIGNAL_UNLOCK ();
       return 0;
     }
+  if (node->flags & G_SIGNAL_NO_HOOKS) 
+    {
+      g_warning ("%s: signal id `%u' does not support emission hooks (G_SIGNAL_NO_HOOKS flag set)", G_STRLOC, signal_id);
+      SIGNAL_UNLOCK ();
+      return 0;
+    }
   if (detail && !(node->flags & G_SIGNAL_DETAILED))
     {
       g_warning ("%s: signal id `%u' does not support detail (%u)", G_STRLOC, signal_id, detail);
@@ -829,7 +897,7 @@ g_signal_add_emission_hook (guint               signal_id,
     }
   hook = g_hook_alloc (node->emission_hooks);
   hook->data = hook_data;
-  hook->func = hook_func;
+  hook->func = (gpointer) hook_func;
   hook->destroy = data_destroy;
   signal_hook = SIGNAL_HOOK (hook);
   signal_hook->detail = detail;
@@ -984,7 +1052,6 @@ g_signal_lookup (const gchar *name,
                  GType        itype)
 {
   guint signal_id;
-  
   g_return_val_if_fail (name != NULL, 0);
   g_return_val_if_fail (G_TYPE_IS_INSTANTIATABLE (itype) || G_TYPE_IS_INTERFACE (itype), 0);
   
@@ -1021,8 +1088,8 @@ g_signal_list_ids (GType  itype,
   g_return_val_if_fail (n_ids != NULL, NULL);
   
   SIGNAL_LOCK ();
-  keys = G_BSEARCH_ARRAY_NODES (g_signal_key_bsa);
-  n_nodes = g_signal_key_bsa->n_nodes;
+  keys = g_bsearch_array_get_nth (g_signal_key_bsa, &g_signal_key_bconfig, 0);
+  n_nodes = g_bsearch_array_get_n_nodes (g_signal_key_bsa);
   result = g_array_new (FALSE, FALSE, sizeof (guint));
   
   for (i = 0; i < n_nodes; i++)
@@ -1119,6 +1186,18 @@ g_signal_new (const gchar	 *signal_name,
                                    return_type, n_params, args);
 
   va_end (args);
+
+  /* optimize NOP emissions with NULL class handlers */
+  if (signal_id && G_TYPE_IS_INSTANTIATABLE (itype) && return_type == G_TYPE_NONE &&
+      class_offset && class_offset < MAX_TEST_CLASS_OFFSET)
+    {
+      SignalNode *node;
+
+      SIGNAL_LOCK ();
+      node = LOOKUP_SIGNAL_NODE (signal_id);
+      node->test_class_offset = class_offset;
+      SIGNAL_UNLOCK ();
+    }
  
   return signal_id;
 }
@@ -1155,8 +1234,8 @@ signal_lookup_closure (SignalNode    *node,
 {
   ClassClosure *cc;
 
-  if (node->class_closure_bsa && node->class_closure_bsa->n_nodes == 1)
-    cc = G_BSEARCH_ARRAY_NODES (node->class_closure_bsa);
+  if (node->class_closure_bsa && g_bsearch_array_get_n_nodes (node->class_closure_bsa) == 1)
+    cc = g_bsearch_array_get_nth (node->class_closure_bsa, &g_class_closure_bconfig, 0);
   else
     cc = signal_find_class_closure (node, G_TYPE_FROM_INSTANCE (instance));
   return cc ? cc->closure : NULL;
@@ -1169,14 +1248,16 @@ signal_add_class_closure (SignalNode *node,
 {
   ClassClosure key;
 
+  /* can't optimize NOP emissions with overridden class closures */
+  node->test_class_offset = 0;
+
   if (!node->class_closure_bsa)
-    node->class_closure_bsa = g_bsearch_array_new (&g_class_closure_bconfig);
+    node->class_closure_bsa = g_bsearch_array_create (&g_class_closure_bconfig);
   key.instance_type = itype;
   key.closure = g_closure_ref (closure);
   node->class_closure_bsa = g_bsearch_array_insert (node->class_closure_bsa,
 						    &g_class_closure_bconfig,
-						    &key,
-						    FALSE);
+						    &key);
   g_closure_sink (closure);
   if (node->c_marshaller && closure && G_CLOSURE_NEEDS_MARSHAL (closure))
     g_closure_set_marshal (closure, node->c_marshaller);
@@ -1219,7 +1300,7 @@ g_signal_newv (const gchar       *signal_name,
     {
       g_warning (G_STRLOC ": signal \"%s\" already exists in the `%s' %s",
                  name,
-                 g_type_name (node->itype),
+                 type_debug_name (node->itype),
                  G_TYPE_IS_INTERFACE (node->itype) ? "interface" : "class ancestry");
       g_free (name);
       SIGNAL_UNLOCK ();
@@ -1229,8 +1310,8 @@ g_signal_newv (const gchar       *signal_name,
     {
       g_warning (G_STRLOC ": signal \"%s\" for type `%s' was previously created for type `%s'",
                  name,
-                 g_type_name (itype),
-                 g_type_name (node->itype));
+                 type_debug_name (itype),
+                 type_debug_name (node->itype));
       g_free (name);
       SIGNAL_UNLOCK ();
       return 0;
@@ -1239,7 +1320,7 @@ g_signal_newv (const gchar       *signal_name,
     if (!G_TYPE_IS_VALUE (param_types[i] & ~G_SIGNAL_TYPE_STATIC_SCOPE))
       {
 	g_warning (G_STRLOC ": parameter %d of type `%s' for signal \"%s::%s\" is not a value type",
-		   i + 1, g_type_name (param_types[i] & ~G_SIGNAL_TYPE_STATIC_SCOPE), g_type_name (itype), name);
+		   i + 1, type_debug_name (param_types[i]), type_debug_name (itype), name);
 	g_free (name);
 	SIGNAL_UNLOCK ();
 	return 0;
@@ -1247,7 +1328,7 @@ g_signal_newv (const gchar       *signal_name,
   if (return_type != G_TYPE_NONE && !G_TYPE_IS_VALUE (return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE))
     {
       g_warning (G_STRLOC ": return value of type `%s' for signal \"%s::%s\" is not a value type",
-		 g_type_name (return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE), g_type_name (itype), name);
+		 type_debug_name (return_type), type_debug_name (itype), name);
       g_free (name);
       SIGNAL_UNLOCK ();
       return 0;
@@ -1256,8 +1337,7 @@ g_signal_newv (const gchar       *signal_name,
       (signal_flags & (G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST | G_SIGNAL_RUN_CLEANUP)) == G_SIGNAL_RUN_FIRST)
     {
       g_warning (G_STRLOC ": signal \"%s::%s\" has return type `%s' and is only G_SIGNAL_RUN_FIRST",
-		 g_type_name (itype), name,
-		 g_type_name (return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE));
+		 type_debug_name (itype), name, type_debug_name (return_type));
       g_free (name);
       SIGNAL_UNLOCK ();
       return 0;
@@ -1278,13 +1358,14 @@ g_signal_newv (const gchar       *signal_name,
       key.itype = itype;
       key.quark = g_quark_from_string (node->name);
       key.signal_id = signal_id;
-      g_signal_key_bsa = g_bsearch_array_insert (g_signal_key_bsa, &g_signal_key_bconfig, &key, FALSE);
+      g_signal_key_bsa = g_bsearch_array_insert (g_signal_key_bsa, &g_signal_key_bconfig, &key);
       g_strdelimit (node->name, "_", '-');
       key.quark = g_quark_from_static_string (node->name);
-      g_signal_key_bsa = g_bsearch_array_insert (g_signal_key_bsa, &g_signal_key_bconfig, &key, FALSE);
+      g_signal_key_bsa = g_bsearch_array_insert (g_signal_key_bsa, &g_signal_key_bconfig, &key);
     }
   node->destroyed = FALSE;
-  
+  node->test_class_offset = 0;
+
   /* setup reinitializable portion */
   node->flags = signal_flags & G_SIGNAL_FLAGS_MASK;
   node->n_params = n_params;
@@ -1303,6 +1384,11 @@ g_signal_newv (const gchar       *signal_name,
   node->emission_hooks = NULL;
   if (class_closure)
     signal_add_class_closure (node, 0, class_closure);
+  else if (G_TYPE_IS_INSTANTIATABLE (itype) && return_type == G_TYPE_NONE)
+    {
+      /* optimize NOP emissions */
+      node->test_class_offset = TEST_CLASS_MAGIC;
+    }
   SIGNAL_UNLOCK ();
 
   return signal_id;
@@ -1350,6 +1436,7 @@ signal_destroy_R (SignalNode *signal_node)
   signal_node->destroyed = TRUE;
   
   /* reentrancy caution, zero out real contents first */
+  signal_node->test_class_offset = 0;
   signal_node->n_params = 0;
   signal_node->param_types = NULL;
   signal_node->return_type = 0;
@@ -1385,7 +1472,7 @@ signal_destroy_R (SignalNode *signal_node)
 
 	  g_closure_unref (cc->closure);
 	}
-      g_bsearch_array_destroy (node.class_closure_bsa, &g_class_closure_bconfig);
+      g_bsearch_array_free (node.class_closure_bsa, &g_class_closure_bconfig);
     }
   g_free (node.accumulator);
   if (node.emission_hooks)
@@ -1409,13 +1496,13 @@ g_signal_override_class_closure (guint     signal_id,
   SIGNAL_LOCK ();
   node = LOOKUP_SIGNAL_NODE (signal_id);
   if (!g_type_is_a (instance_type, node->itype))
-    g_warning ("%s: type `%s' cannot be overridden for signal id `%u'", G_STRLOC, g_type_name (instance_type), signal_id);
+    g_warning ("%s: type `%s' cannot be overridden for signal id `%u'", G_STRLOC, type_debug_name (instance_type), signal_id);
   else
     {
       ClassClosure *cc = signal_find_class_closure (node, instance_type);
       
       if (cc && cc->instance_type == instance_type)
-	g_warning ("%s: type `%s' is already overridden for signal id `%u'", G_STRLOC, g_type_name (instance_type), signal_id);
+	g_warning ("%s: type `%s' is already overridden for signal id `%u'", G_STRLOC, type_debug_name (instance_type), signal_id);
       else
 	signal_add_class_closure (node, instance_type, class_closure);
     }
@@ -1713,7 +1800,6 @@ g_signal_handler_is_connected (gpointer instance,
   gboolean connected;
 
   g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), FALSE);
-  g_return_val_if_fail (handler_id > 0, FALSE);
 
   SIGNAL_LOCK ();
   handler = handler_lookup (instance, handler_id, NULL);
@@ -1760,7 +1846,7 @@ g_signal_handlers_destroy (gpointer instance)
 		}
             }
         }
-      g_bsearch_array_destroy (hlbsa, &g_signal_hlbsa_bconfig);
+      g_bsearch_array_free (hlbsa, &g_signal_hlbsa_bconfig);
     }
   SIGNAL_UNLOCK ();
 }
@@ -1837,8 +1923,8 @@ g_signal_handlers_block_matched (gpointer         instance,
 {
   guint n_handlers = 0;
   
-  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), FALSE);
-  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, FALSE);
+  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), 0);
+  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, 0);
   
   if (mask & (G_SIGNAL_MATCH_CLOSURE | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA))
     {
@@ -1863,8 +1949,8 @@ g_signal_handlers_unblock_matched (gpointer         instance,
 {
   guint n_handlers = 0;
   
-  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), FALSE);
-  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, FALSE);
+  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), 0);
+  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, 0);
   
   if (mask & (G_SIGNAL_MATCH_CLOSURE | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA))
     {
@@ -1889,8 +1975,8 @@ g_signal_handlers_disconnect_matched (gpointer         instance,
 {
   guint n_handlers = 0;
   
-  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), FALSE);
-  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, FALSE);
+  g_return_val_if_fail (G_TYPE_CHECK_INSTANCE (instance), 0);
+  g_return_val_if_fail ((mask & ~G_SIGNAL_MATCH_MASK) == 0, 0);
   
   if (mask & (G_SIGNAL_MATCH_CLOSURE | G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA))
     {
@@ -1943,6 +2029,50 @@ g_signal_has_handler_pending (gpointer instance,
   return has_pending;
 }
 
+static inline gboolean
+signal_check_skip_emission (SignalNode *node,
+			    gpointer    instance,
+			    GQuark      detail)
+{
+  HandlerList *hlist;
+
+  /* are we able to check for NULL class handlers? */
+  if (!node->test_class_offset)
+    return FALSE;
+
+  /* are there emission hooks pending? */
+  if (node->emission_hooks && node->emission_hooks->hooks)
+    return FALSE;
+
+  /* is there a non-NULL class handler? */
+  if (node->test_class_offset != TEST_CLASS_MAGIC)
+    {
+      GTypeClass *class = G_TYPE_INSTANCE_GET_CLASS (instance, G_TYPE_FROM_INSTANCE (instance), GTypeClass);
+
+      if (G_STRUCT_MEMBER (gpointer, class, node->test_class_offset))
+	return FALSE;
+    }
+
+  /* are signals being debugged? */
+#ifdef  G_ENABLE_DEBUG
+  IF_DEBUG (SIGNALS, g_trace_instance_signals || g_trap_instance_signals)
+    return FALSE;
+#endif /* G_ENABLE_DEBUG */
+
+  /* is this a no-recurse signal already in emission? */
+  if (node->flags & G_SIGNAL_NO_RECURSE &&
+      emission_find (g_restart_emissions, node->signal_id, detail, instance))
+    return FALSE;
+
+  /* do we have pending handlers? */
+  hlist = handler_list_lookup (node->signal_id, instance);
+  if (hlist && hlist->handlers)
+    return FALSE;
+
+  /* none of the above, no emission required */
+  return TRUE;
+}
+
 void
 g_signal_emitv (const GValue *instance_and_params,
 		guint         signal_id,
@@ -1983,7 +2113,7 @@ g_signal_emitv (const GValue *instance_and_params,
       {
 	g_critical ("%s: value for `%s' parameter %u for signal \"%s\" is of type `%s'",
 		    G_STRLOC,
-		    g_type_name (node->param_types[i] & ~G_SIGNAL_TYPE_STATIC_SCOPE),
+		    type_debug_name (node->param_types[i]),
 		    i,
 		    node->name,
 		    G_VALUE_TYPE_NAME (param_values + i));
@@ -1996,7 +2126,7 @@ g_signal_emitv (const GValue *instance_and_params,
 	{
 	  g_critical ("%s: return value `%s' for signal \"%s\" is (NULL)",
 		      G_STRLOC,
-		      g_type_name (node->return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE),
+		      type_debug_name (node->return_type),
 		      node->name);
 	  SIGNAL_UNLOCK ();
 	  return;
@@ -2005,7 +2135,7 @@ g_signal_emitv (const GValue *instance_and_params,
 	{
 	  g_critical ("%s: return value `%s' for signal \"%s\" is of type `%s'",
 		      G_STRLOC,
-		      g_type_name (node->return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE),
+		      type_debug_name (node->return_type),
 		      node->name,
 		      G_VALUE_TYPE_NAME (return_value));
 	  SIGNAL_UNLOCK ();
@@ -2016,6 +2146,15 @@ g_signal_emitv (const GValue *instance_and_params,
     return_value = NULL;
 #endif	/* G_ENABLE_DEBUG */
 
+  /* optimize NOP emissions */
+  if (signal_check_skip_emission (node, instance, detail))
+    {
+      /* nothing to do to emit this signal */
+      SIGNAL_UNLOCK ();
+      /* g_printerr ("omitting emission of \"%s\"\n", node->name); */
+      return;
+    }
+
   SIGNAL_UNLOCK ();
   signal_emit_unlocked_R (node, detail, instance, return_value, instance_and_params);
 }
@@ -2052,6 +2191,15 @@ g_signal_emit_valist (gpointer instance,
     }
 #endif  /* !G_DISABLE_CHECKS */
 
+  /* optimize NOP emissions */
+  if (signal_check_skip_emission (node, instance, detail))
+    {
+      /* nothing to do to emit this signal */
+      SIGNAL_UNLOCK ();
+      /* g_printerr ("omitting emission of \"%s\"\n", node->name); */
+      return;
+    }
+
   n_params = node->n_params;
   signal_return_type = node->return_type;
   if (node->n_params < MAX_STACK_VALUES)
@@ -2120,7 +2268,7 @@ g_signal_emit_valist (gpointer instance,
 	  g_free (error);
 	  
 	  /* we purposely leak the value here, it might not be
-	   * in a sane state if an error condition occoured
+	   * in a sane state if an error condition occured
 	   */
 	}
     }
@@ -2298,7 +2446,7 @@ signal_emit_unlocked_R (SignalNode   *node,
 	  
 	  if (!signal_hook->detail || signal_hook->detail == detail)
 	    {
-	      GSignalEmissionHook hook_func = hook->func;
+	      GSignalEmissionHook hook_func = (GSignalEmissionHook) hook->func;
 	      
 	      was_in_call = G_HOOK_IN_CALL (hook);
 	      hook->flags |= G_HOOK_FLAG_IN_CALL;
@@ -2478,8 +2626,38 @@ signal_emit_unlocked_R (SignalNode   *node,
   return return_value_altered;
 }
 
+static const gchar*
+type_debug_name (GType type)
+{
+  if (type)
+    {
+      const char *name = g_type_name (type & ~G_SIGNAL_TYPE_STATIC_SCOPE);
+      return name ? name : "<unknown>";
+    }
+  else
+    return "<invalid>";
+}
+
+gboolean
+g_signal_accumulator_true_handled (GSignalInvocationHint *ihint,
+				   GValue                *return_accu,
+				   const GValue          *handler_return,
+				   gpointer               dummy)
+{
+  gboolean continue_emission;
+  gboolean signal_handled;
+  
+  signal_handled = g_value_get_boolean (handler_return);
+  g_value_set_boolean (return_accu, signal_handled);
+  continue_emission = !signal_handled;
+  
+  return continue_emission;
+}
 
 /* --- compile standard marshallers --- */
 #include	"gobject.h"
 #include	"genums.h"
 #include        "gmarshal.c"
+
+#define __G_SIGNAL_C__
+#include "gobjectaliasdef.c"
