@@ -24,6 +24,7 @@
 #include	"gvaluetypes.h"
 
 #include	"gvaluecollector.h"
+#include	"gobjectalias.h"
 #include	<string.h>
 #include	<stdlib.h>	/* qsort() */
 
@@ -357,7 +358,7 @@ value_lcopy_pointer (const GValue *value,
 
 /* --- type initialization --- */
 void
-g_value_types_init (void)  /* sync with gtype.c */
+g_value_types_init (void)
 {
   GTypeInfo info = {
     0,				/* class_size */
@@ -762,6 +763,13 @@ g_value_set_static_string (GValue      *value,
 void
 g_value_set_string_take_ownership (GValue *value,
 				   gchar  *v_string)
+{
+  g_value_take_string (value, v_string);
+}
+
+void
+g_value_take_string (GValue *value,
+		     gchar  *v_string)
 {
   g_return_if_fail (G_VALUE_HOLDS_STRING (value));
   
@@ -898,3 +906,5 @@ g_pointer_type_register_static (const gchar *name)
   return type;
 }
 
+#define __G_VALUETYPES_C__
+#include "gobjectaliasdef.c"
