@@ -50,12 +50,14 @@
 static int array[10000];
 static gboolean failed = FALSE;
 
+/* We write (m ? m : "") even in the m != NULL case to suppress a warning with GCC-3.1
+ */
 #define	TEST(m,cond)	G_STMT_START { failed = !(cond); \
 if (failed) \
   { if (!m) \
       g_print ("\n(%s:%d) failed for: %s\n", __FILE__, __LINE__, ( # cond )); \
     else \
-      g_print ("\n(%s:%d) failed for: %s: (%s)\n", __FILE__, __LINE__, ( # cond ), (gchar*)m); \
+      g_print ("\n(%s:%d) failed for: %s: (%s)\n", __FILE__, __LINE__, ( # cond ), (gchar*)(m ? m : "")); \
   } \
 else \
   g_print ("."); fflush (stdout); \
@@ -305,26 +307,44 @@ my_traverse (gpointer key,
   return FALSE;
 }
 
+static gboolean 
+find_first_that(gpointer key, 
+		gpointer value, 
+		gpointer user_data)
+{
+  gint *v = value;
+  gint *test = user_data;
+  return (*v == *test);
+}
+
+
 int
 main (int   argc,
       char *argv[])
 {
+  const gchar *s;
+  gchar **sv;
   GList *list, *t;
   GSList *slist, *st;
   GHashTable *hash_table;
   GMemChunk *mem_chunk;
   GStringChunk *string_chunk;
-  GTimer *timer;
+  GTimer *timer, *timer2;
+  gdouble elapsed;
+  gulong elapsed_usecs;
   gint nums[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
   gint morenums[10] = { 8, 9, 7, 0, 3, 2, 5, 1, 4, 6};
   gchar *string;
-
+  gint value = 120; 
+  gint *pvalue=NULL; 
+  
   gchar *mem[10000], *tmp_string = NULL, *tmp_string_2;
   gint i, j;
   GArray *garray;
   GPtrArray *gparray;
   GByteArray *gbarray;
   GString *string1, *string2;
+  const gchar *charset;
   GTree *tree;
   char chars[62];
   GRelation *relation;
@@ -334,7 +354,6 @@ main (int   argc,
     gchar *filename;
     gchar *dirname;
   } dirname_checks[] = {
-#ifndef G_OS_WIN32
     { "/", "/" },
     { "////", "/" },
     { ".////", "." },
@@ -343,14 +362,16 @@ main (int   argc,
     { "a/b", "a" },
     { "a/b/", "a/b" },
     { "c///", "c" },
-#else
+#ifdef G_OS_WIN32
     { "\\", "\\" },
     { ".\\\\\\\\", "." },
     { "..\\", ".." },
     { "..\\\\\\\\", ".." },
     { "a\\b", "a" },
-    { "a\\b\\", "a\\b" },
-    { "c\\\\\\", "c" },
+    { "a\\b/", "a\\b" },
+    { "a/b\\", "a/b" },
+    { "c\\\\/", "c" },
+    { "//\\", "/" },
 #endif
 #ifdef G_WITH_CYGWIN
     { "//server/share///x", "//server/share" },
@@ -365,13 +386,12 @@ main (int   argc,
     gchar *filename;
     gchar *without_root;
   } skip_root_checks[] = {
-#ifndef G_OS_WIN32
     { "/", "" },
     { "//", "" },
     { "/foo", "foo" },
     { "//foo", "foo" },
     { "a/b", NULL },
-#else
+#ifdef G_OS_WIN32
     { "\\", "" },
     { "\\foo", "foo" },
     { "\\\\server\\foo", "" },
@@ -389,10 +409,8 @@ main (int   argc,
 #ifndef G_DISABLE_ASSERT
   guint16 gu16t1 = 0x44afU, gu16t2 = 0xaf44U;
   guint32 gu32t1 = 0x02a7f109U, gu32t2 = 0x09f1a702U;
-#ifdef G_HAVE_GINT64
   guint64 gu64t1 = G_GINT64_CONSTANT(0x1d636b02300a7aa7U),
 	  gu64t2 = G_GINT64_CONSTANT(0xa77a0a30026b631dU);
-#endif
 #endif
   const char hello[] = "Hello, World";
   const int hellolen = sizeof (hello) - 1;
@@ -401,16 +419,11 @@ main (int   argc,
   GError *error;
   char *name_used;
 #ifdef G_OS_WIN32
-  gchar *glib_dll = g_strdup_printf ("libglib-%d.%d-%d.dll",
-				     GLIB_MAJOR_VERSION,
-				     GLIB_MINOR_VERSION,
-				     GLIB_MICRO_VERSION - GLIB_BINARY_AGE);
+  /* Can't calculate GLib DLL name at runtime. */
+  gchar *glib_dll = "libglib-2.0-0.dll";
 #endif
 #ifdef G_WITH_CYGWIN
-  gchar *glib_dll = g_strdup_printf ("cygglib-%d.%d-%d.dll",
-				     GLIB_MAJOR_VERSION,
-				     GLIB_MINOR_VERSION,
-				     GLIB_MICRO_VERSION - GLIB_BINARY_AGE);
+  gchar *glib_dll = "cygglib-2.0-0.dll";
 #endif
 
   g_print ("TestGLib v%u.%u.%u (i:%u b:%u)\n",
@@ -425,21 +438,32 @@ main (int   argc,
   g_free (string);
   g_print ("user: %s\n", g_get_user_name ());
   g_print ("real: %s\n", g_get_real_name ());
-  g_print ("home: %s\n", g_get_home_dir ());
+  s = g_get_home_dir ();
+  g_print ("home: %s\n", s ? s : "NULL!");
+  s = g_get_user_data_dir ();
+  g_print ("user_data: %s\n", s ? s : "NULL!");
+  s = g_get_user_config_dir ();
+  g_print ("user_config: %s\n", s ? s : "NULL!");
+  s = g_get_user_cache_dir ();
+  g_print ("user_cache: %s\n", s ? s : "NULL!");
+  sv = (gchar **) g_get_system_data_dirs ();
+  g_print ("system_data: %s\n", sv ? g_strjoinv (G_SEARCHPATH_SEPARATOR_S, sv) : "NULL!");
+  sv = (gchar **) g_get_system_config_dirs ();
+  g_print ("system_config: %s\n", s ? g_strjoinv (G_SEARCHPATH_SEPARATOR_S, sv) : "NULL!");
   g_print ("tmp-dir: %s\n", g_get_tmp_dir ());
+  sv = (gchar **) g_get_language_names ();
+  g_print ("languages: %s\n", s ? g_strjoinv (":", sv) : "NULL!");
 
   /* type sizes */
-  g_print ("checking size of gint8: %d", (int)sizeof (gint8));
+  g_print ("checking size of gint8: %"    G_GSIZE_FORMAT, sizeof (gint8));
   TEST (NULL, sizeof (gint8) == 1);
-  g_print ("\nchecking size of gint16: %d", (int)sizeof (gint16));
+  g_print ("\nchecking size of gint16: %" G_GSIZE_FORMAT, sizeof (gint16));
   TEST (NULL, sizeof (gint16) == 2);
-  g_print ("\nchecking size of gint32: %d", (int)sizeof (gint32));
+  g_print ("\nchecking size of gint32: %" G_GSIZE_FORMAT, sizeof (gint32));
   TEST (NULL, sizeof (gint32) == 4);
-  g_print ("\nchecking size of gsize: %d", (int)sizeof (gsize));
-#ifdef	G_HAVE_GINT64
-  g_print ("\nchecking size of gint64: %d", (int)sizeof (gint64));
+  g_print ("\nchecking size of gsize: %"  G_GSIZE_FORMAT, sizeof (gsize));
+  g_print ("\nchecking size of gint64: %" G_GSIZE_FORMAT, sizeof (gint64));
   TEST (NULL, sizeof (gint64) == 8);
-#endif	/* G_HAVE_GINT64 */
   g_print ("\n");
 
   g_print ("checking g_path_get_basename()...");
@@ -450,6 +474,15 @@ main (int   argc,
   g_assert (strcmp (string, "file") == 0);
   g_free (string);
   g_print ("ok\n");
+#ifdef G_OS_WIN32
+  string = g_path_get_basename ("/foo/dir/");
+  g_assert (strcmp (string, "dir") == 0);
+  g_free (string);
+  string = g_path_get_basename ("/foo/file");
+  g_assert (strcmp (string, "file") == 0);
+  g_free (string);
+  g_print ("ok\n");
+#endif
 
   g_print ("checking g_path_get_dirname()...");
   for (i = 0; i < n_dirname_checks; i++)
@@ -670,7 +703,7 @@ main (int   argc,
   g_print ("tree nnodes: %d\n", g_tree_nnodes (tree));
 
   g_print ("tree: ");
-  g_tree_traverse (tree, my_traverse, G_IN_ORDER, NULL);
+  g_tree_foreach (tree, my_traverse, NULL);
   g_print ("\n");
 
   for (i = 0; i < 10; i++)
@@ -680,7 +713,7 @@ main (int   argc,
   g_print ("tree nnodes: %d\n", g_tree_nnodes (tree));
 
   g_print ("tree: ");
-  g_tree_traverse (tree, my_traverse, G_IN_ORDER, NULL);
+  g_tree_foreach (tree, my_traverse, NULL);
   g_print ("\n");
 
   g_print ("ok\n");
@@ -717,6 +750,10 @@ main (int   argc,
       array[i] = i;
       g_hash_table_insert (hash_table, &array[i], &array[i]);
     }
+  pvalue = g_hash_table_find (hash_table, find_first_that, &value);
+  if (*pvalue != value)
+	  g_print("g_hash_table_find failed");
+  
   g_hash_table_foreach (hash_table, my_hash_callback, NULL);
 
   for (i = 0; i < 10000; i++)
@@ -808,19 +845,19 @@ main (int   argc,
 #ifndef G_OS_WIN32
   /* MSVC, mingw32 and LCC use the same run-time C library, which doesn't like
      the %10000.10000f format... */
-  g_string_sprintf (string2, "%s|%0100d|%s|%s|%0*d|%*.*f|%10000.10000f",
-		    "this pete guy sure is a wuss, like he's the number ",
-		    1,
-		    " wuss.  everyone agrees.\n",
-		    string1->str,
-		    10, 666, 15, 15, 666.666666666, 666.666666666);
+  g_string_printf (string2, "%s|%0100d|%s|%s|%0*d|%*.*f|%10000.10000f",
+		   "this pete guy sure is a wuss, like he's the number ",
+		   1,
+		   " wuss.  everyone agrees.\n",
+		   string1->str,
+		   10, 666, 15, 15, 666.666666666, 666.666666666);
 #else
-  g_string_sprintf (string2, "%s|%0100d|%s|%s|%0*d|%*.*f|%100.100f",
-		    "this pete guy sure is a wuss, like he's the number ",
-		    1,
-		    " wuss.  everyone agrees.\n",
-		    string1->str,
-		    10, 666, 15, 15, 666.666666666, 666.666666666);
+  g_string_printf (string2, "%s|%0100d|%s|%s|%0*d|%*.*f|%100.100f",
+		   "this pete guy sure is a wuss, like he's the number ",
+		   1,
+		   " wuss.  everyone agrees.\n",
+		   string1->str,
+		   10, 666, 15, 15, 666.666666666, 666.666666666);
 #endif
 
   g_print ("string2 length = %lu...\n", (gulong)string2->len);
@@ -943,6 +980,60 @@ main (int   argc,
 
   g_print ("ok\n");
 
+  g_print ("checking g_timer_continue...\n");
+
+  timer2 = g_timer_new ();
+
+  g_print ("\trun for 1 second...\n");
+  timer = g_timer_new();
+  g_usleep(G_USEC_PER_SEC); /* run timer for 1 second */
+  g_timer_stop(timer);
+
+  g_print ("\tstop for 1 second...\n");
+  g_usleep(G_USEC_PER_SEC); /* wait for 1 second */
+  g_print ("\trun for 2 seconds...\n");
+
+  g_timer_continue(timer);
+  g_usleep(2*G_USEC_PER_SEC); /* run timer for 2 seconds */
+  g_timer_stop(timer);
+
+  g_print ("\tstop for 1.5 seconds...\n");
+  g_usleep((3*G_USEC_PER_SEC)/2); /* wait for 1.5 seconds */
+  g_print ("\trun for 0.2 seconds...\n");
+
+  g_timer_continue(timer);
+  g_usleep(G_USEC_PER_SEC/5); /* run timer for 0.2 seconds */
+  g_timer_stop(timer);
+
+  g_print ("\tstop for 4 seconds...\n");
+  g_usleep(4*G_USEC_PER_SEC); /* wait for 4 seconds */
+  g_print ("\trun for 5.8 seconds...\n");
+
+  g_timer_continue(timer);
+  g_usleep((29*G_USEC_PER_SEC)/5); /* run timer for 5.8 seconds */
+  g_timer_stop(timer);
+
+  elapsed = g_timer_elapsed (timer, &elapsed_usecs);
+  g_print ("\t=> timer = %.6f = %d.%06ld (should be: 9.000000) (%.6f off)\n", elapsed, (int) elapsed, elapsed_usecs, ABS (elapsed - 9.));
+
+  if (elapsed > 8.8 && elapsed < 9.2)
+    g_print ("g_timer_continue ... ok\n\n");
+  else
+    g_print ("g_timer_continue ... ***** FAILED *****\n\n");
+
+  g_timer_stop(timer2);
+
+  elapsed = g_timer_elapsed(timer2, &elapsed_usecs);
+  g_print ("\t=> timer2 = %.6f = %d.%06ld (should be: %.6f) (%.6f off)\n\n", elapsed, (int) elapsed, elapsed_usecs, 9.+6.5, ABS (elapsed - (9.+6.5)));
+
+  if (elapsed > (8.8+6.5) && elapsed < (9.2+6.5))
+    g_print ("timer2 ... ok\n\n");
+  else
+    g_print ("timer2 ... ***** FAILED *****\n\n");
+
+  g_timer_destroy(timer);
+  g_timer_destroy(timer2);
+
   g_print ("checking g_ascii_strcasecmp...");
   g_assert (g_ascii_strcasecmp ("FroboZZ", "frobozz") == 0);
   g_assert (g_ascii_strcasecmp ("frobozz", "frobozz") == 0);
@@ -1160,6 +1251,8 @@ main (int   argc,
   g_message ("the next warning is a test:");
   string = NULL;
   g_print (string);
+  g_message ("non-printable UTF-8: \"\xc3\xa4\xda\x85\"");
+  g_message ("unsafe chars: \"\x10\x11\x12\n\t\x7f\x81\x82\x83\"");
 
   g_print ("checking endian macros (host is ");
 #if G_BYTE_ORDER == G_BIG_ENDIAN
@@ -1169,12 +1262,15 @@ main (int   argc,
 #endif
   g_assert (GUINT16_SWAP_LE_BE (gu16t1) == gu16t2);  
   g_assert (GUINT32_SWAP_LE_BE (gu32t1) == gu32t2);  
-#ifdef G_HAVE_GINT64
   g_assert (GUINT64_SWAP_LE_BE (gu64t1) == gu64t2);  
-#endif
 
   g_print ("ok\n");
 
+  if (g_get_charset ((G_CONST_RETURN char**)&charset))
+    g_print ("current charset is UTF-8: %s\n", charset);
+  else
+    g_print ("current charset is not UTF-8: %s\n", charset);
+
 #ifdef G_PLATFORM_WIN32
   g_print ("current locale: %s\n", g_win32_getlocale ());
   g_print ("GLib DLL name tested for: %s\n", glib_dll);
@@ -1187,13 +1283,15 @@ main (int   argc,
   g_print ("Ditto, only from GLib DLL name: %s\n",
 	   g_win32_get_package_installation_directory (NULL, glib_dll));
   g_print ("locale subdirectory of GLib installation directory: %s\n",
-	   g_win32_get_package_installation_subdirectory (NULL, glib_dll, "share\\locale"));
+	   g_win32_get_package_installation_subdirectory (NULL, glib_dll, "lib\\locale"));
   g_print ("GTK+ 2.0 installation directory, if available: %s\n",
 	   g_win32_get_package_installation_directory ("gtk20", NULL));
 
   g_print ("found more.com as %s\n", g_find_program_in_path ("more.com"));
   g_print ("found regedit as %s\n", g_find_program_in_path ("regedit"));
 
+  g_print ("a Win32 error message: %s\n", g_win32_error_message (2));
+
 #endif
 
   g_print ("checking file functions...\n");
@@ -1239,6 +1337,18 @@ main (int   argc,
   close (fd);
   g_clear_error (&error);
 
+#ifdef G_OS_WIN32
+  strcpy (template, "zap/barXXXXXX");
+  fd = g_file_open_tmp (template, &name_used, &error);
+  if (fd != -1)
+    g_print ("g_file_open_tmp works even if template contains '/'\n");
+  else
+    g_print ("g_file_open_tmp correctly returns error: %s\n",
+	     error->message);
+  close (fd);
+  g_clear_error (&error);
+#endif
+
   strcpy (template, "zapXXXXXX");
   fd = g_file_open_tmp (template, &name_used, &error);
   if (fd == -1)
